@@ -21,26 +21,29 @@ Serpent handles several special Python types to make life easier:
  - Exception  --> dict with some fields of the exception (message, args)
  - all other types  --> dict with  __getstate__  or vars() of the object
 
-Note: all str will be promoted to unicode. This is done because it is the
+Notes:
+
+All str will be promoted to unicode. This is done because it is the
 default anyway for Python 3.x, and it solves the problem of the str/unicode
 difference between different Python versions. Also it means the serialized
 output doesn't have those problematic 'u' prefixes on strings.
 
-Note: the serializer is not thread-safe. Make sure you're not making changes
+The serializer is not thread-safe. Make sure you're not making changes
 to the object tree that is being serialized, and don't use the same
 serializer in different threads.
 
-Caveat: Python 2.6 cannot deserialize complex numbers (limitation of
+Python 2.6 cannot deserialize complex numbers (limitation of
 ast.literal_eval in 2.6)
 
-Note: because the serialized format is just valid Python source code, it can
+Because the serialized format is just valid Python source code, it can
 contain comments.
 
-Note: set literals are not supported on python <3.2 (ast.literal_eval
+Set literals are not supported on python <3.2 (ast.literal_eval
 limitation). If you need Python < 3.2 compatibility, you'll have to use
-set_literals=False when serializing.
+set_literals=False when serializing. Since version 1.6 serpent chooses
+this wisely for you by default, but you can still override it if needed.
 
-Copyright 2013, Irmen de Jong (irmen@razorvine.net)
+Copyright 2013, 2014 by Irmen de Jong (irmen@razorvine.net)
 Software license: "MIT software license". See http://opensource.org/licenses/MIT
 """
 
@@ -53,18 +56,20 @@ import types
 import os
 import gc
 
-__version__ = "1.4"
+__version__ = "1.8"
 __all__ = ["dump", "dumps", "load", "loads", "register_class", "unregister_class"]
 
+can_use_set_literals = sys.version_info >= (3, 2)  # check if we can use set literals
 
-def dumps(obj, indent=False, set_literals=True):
+
+def dumps(obj, indent=False, set_literals=can_use_set_literals, module_in_classname=False):
     """Serialize object tree to bytes"""
-    return Serializer(indent, set_literals).serialize(obj)
+    return Serializer(indent, set_literals, module_in_classname).serialize(obj)
 
 
-def dump(obj, file, indent=False, set_literals=True):
+def dump(obj, file, indent=False, set_literals=can_use_set_literals, module_in_classname=False):
     """Serialize object tree to a file"""
-    file.write(dumps(obj, indent=indent, set_literals=set_literals))
+    file.write(dumps(obj, indent=indent, set_literals=set_literals, module_in_classname=module_in_classname))
 
 
 def loads(serialized_bytes):
@@ -105,7 +110,7 @@ def unregister_class(clazz):
 
 def register_class(clazz, serializer):
     """
-    Register a specialcase serializer function for objects of the given class.
+    Register a special serializer function for objects of the given class.
     The function will be called with (object, serpent_serializer, outputstream, indentlevel) arguments.
     The function must write the serialized data to outputstream. It doesn't return a value.
     """
@@ -157,9 +162,10 @@ class Serializer(object):
     """
     Serialize an object tree to a byte stream.
     It is not thread-safe: make sure you're not making changes to the
-    object tree that is being serialized.
+    object tree that is being serialized, and don't use the same serializer
+    across different threads.
     """
-    #noinspection PySetFunctionToLiteral
+    # noinspection PySetFunctionToLiteral
     repr_types = set([
         str,
         int,
@@ -188,17 +194,22 @@ class Serializer(object):
     if sys.version_info < (2, 7):
         repr_types.remove(float)   # repr(float) prints floating point roundoffs in Python < 2.7
 
-    def __init__(self, indent=False, set_literals=True):
+    def __init__(self, indent=False, set_literals=can_use_set_literals, module_in_classname=False):
         """
         Initialize the serializer.
         indent=indent the output over multiple lines (default=false)
-        setLiterals=use set-literals or not (set to False if you need compatibility with Python < 3.2)
+        setLiterals=use set-literals or not (set to False if you need compatibility with Python < 3.2). Serpent chooses a sensible default for you.
+        module_in_classname = include module prefix for class names or only use the class name itself
         """
         self.indent = indent
         self.set_literals = set_literals
+        self.module_in_classname = module_in_classname
+        self.serialized_obj_ids = set()
+        self.special_classes_registry_copy = None
 
     def serialize(self, obj):
         """Serialize the object tree to bytes."""
+        self.special_classes_registry_copy = _special_classes_registry.copy()   # make it thread safe
         header = "# serpent utf-8 "
         if self.set_literals:
             header += "python3.2\n"   # set-literals require python 3.2+ to deserialize (ast.literal_eval limitation)
@@ -208,9 +219,12 @@ class Serializer(object):
         try:
             if os.name != "java" and sys.platform != "cli":
                 gc.disable()
+            self.serialized_obj_ids = set()
             self._serialize(obj, out, 0)
         finally:
             gc.enable()
+        self.special_classes_registry_copy = None
+        del self.serialized_obj_ids
         if sys.platform == "cli":
             return "".join(out)
         return b"".join(out)
@@ -224,9 +238,9 @@ class Serializer(object):
             out.append(_repr(obj))    # just a simple repr() is enough for these objects
             return
         # check special registered types:
-        for clazz in _special_classes_registry:
+        for clazz, class_serializer in self.special_classes_registry_copy.items():
             if isinstance(obj, clazz):
-                _special_classes_registry[clazz](obj, self, out, level)
+                class_serializer(obj, self, out, level)
                 return
         # exception?
         if isinstance(obj, BaseException):
@@ -294,6 +308,9 @@ class Serializer(object):
             out.append(b")")
 
     def ser_builtins_list(self, list_obj, out, level):
+        if id(list_obj) in self.serialized_obj_ids:
+            raise ValueError("Circular reference detected (list)")
+        self.serialized_obj_ids.add(id(list_obj))
         if self.indent and list_obj:
             indent_chars = b"  " * level
             indent_chars_inside = indent_chars + b"  "
@@ -312,8 +329,12 @@ class Serializer(object):
             if list_obj:
                 del out[-1]  # remove the last ,
             out.append(b"]")
+        self.serialized_obj_ids.discard(id(list_obj))
 
     def ser_builtins_dict(self, dict_obj, out, level):
+        if id(dict_obj) in self.serialized_obj_ids:
+            raise ValueError("Circular reference detected (dict)")
+        self.serialized_obj_ids.add(id(dict_obj))
         if self.indent and dict_obj:
             indent_chars = b"  " * level
             indent_chars_inside = indent_chars + b"  "
@@ -341,6 +362,7 @@ class Serializer(object):
             if dict_obj:
                 del out[-1]  # remove the last ,
             out.append(b"}")
+        self.serialized_obj_ids.discard(id(dict_obj))
 
     def ser_builtins_set(self, set_obj, out, level):
         if not self.set_literals:
@@ -385,7 +407,7 @@ class Serializer(object):
 
     if os.name == "java" or sys.version_info < (2, 7):    # jython bug http://bugs.jython.org/issue2010
         def ser_datetime_timedelta(self, timedelta_obj, out, level):
-            secs = ((timedelta_obj.days * 86400 + timedelta_obj.seconds) * 10**6 + timedelta_obj.microseconds) / 10**6
+            secs = ((timedelta_obj.days * 86400 + timedelta_obj.seconds) * 10 ** 6 + timedelta_obj.microseconds) / 10 ** 6
             self._serialize(secs, out, level)
     else:
         def ser_datetime_timedelta(self, timedelta_obj, out, level):
@@ -399,8 +421,12 @@ class Serializer(object):
         self._serialize(str(uuid_obj), out, level)
 
     def ser_exception_class(self, exc_obj, out, level):
+        if self.module_in_classname:
+            class_name = "%s.%s" % (exc_obj.__class__.__module__, exc_obj.__class__.__name__)
+        else:
+            class_name = exc_obj.__class__.__name__
         value = {
-            "__class__": exc_obj.__class__.__name__,
+            "__class__": class_name,
             "__exception__": True,
             "args": exc_obj.args,
             "attributes": vars(exc_obj)  # add any custom attributes
@@ -416,22 +442,32 @@ class Serializer(object):
             self._serialize(array_obj.tolist(), out, level)
 
     def ser_default_class(self, obj, out, level):
+        if id(obj) in self.serialized_obj_ids:
+            raise ValueError("Circular reference detected (class)")
+        self.serialized_obj_ids.add(id(obj))
         try:
-            value = obj.__getstate__()
-            if isinstance(value, dict):
-                self.ser_builtins_dict(value, out, level)
-                return
-        except AttributeError:
             try:
-                value = dict(vars(obj))  # make sure we can serialize anything that resembles a dict
-                value["__class__"] = obj.__class__.__name__
-            except TypeError:
-                if hasattr(obj, "__slots__"):
-                    # use the __slots__ instead of the vars dict
-                    value = {}
-                    for slot in obj.__slots__:
-                        value[slot] = getattr(obj, slot)
-                    value["__class__"] = obj.__class__.__name__
+                value = obj.__getstate__()
+                if isinstance(value, dict):
+                    self.ser_builtins_dict(value, out, level)
+                    return
+            except AttributeError:
+                if self.module_in_classname:
+                    class_name = "%s.%s" % (obj.__class__.__module__, obj.__class__.__name__)
                 else:
-                    raise TypeError("don't know how to serialize class " + str(obj.__class__) + ". Give it vars() or an appropriate __getstate__")
-        self._serialize(value, out, level)
+                    class_name = obj.__class__.__name__
+                try:
+                    value = dict(vars(obj))  # make sure we can serialize anything that resembles a dict
+                    value["__class__"] = class_name
+                except TypeError:
+                    if hasattr(obj, "__slots__"):
+                        # use the __slots__ instead of the vars dict
+                        value = {}
+                        for slot in obj.__slots__:
+                            value[slot] = getattr(obj, slot)
+                        value["__class__"] = class_name
+                    else:
+                        raise TypeError("don't know how to serialize class " + str(obj.__class__) + ". Give it vars() or an appropriate __getstate__")
+            self._serialize(value, out, level)
+        finally:
+            self.serialized_obj_ids.discard(id(obj))
